"""
pytokio implements its command-line tools within this package.  Each such CLI
tool either implements some useful analysis on top of pytokio connectors,
tools, or analysis or maps some of the internal python APIs to command-line
arguments.

Most of these tools implement a ``--help`` option to explain the command-line
options.
"""
